import sys
import os
from .dir_utils import mkdir_p

IRS_READER_ROOT = os.path.abspath(os.path.dirname(__file__))

# This is the URL to amazon's bucket, could use another synced to it
IRS_XML_HTTP_BASE = "https://s3.amazonaws.com/irs-form-990"

# It can be hard to locate this.
IRSX_SETTINGS_LOCATION = (os.path.join(IRS_READER_ROOT, "settings.py"))

# Defaults to the same directory as this settings file, but you can override
# with the `IRSX_CACHE_DIRECTORY` environment variable
IRSX_CACHE_DIRECTORY = os.environ.get("IRSX_CACHE_DIRECTORY", IRS_READER_ROOT)

# The directory we put files in while we're processing them
WORKING_DIRECTORY = os.environ.get(
    "IRSX_WORKING_DIRECTORY", os.path.join(IRSX_CACHE_DIRECTORY, "XML"))
# Helpful to keep these around for lookup purposes
INDEX_DIRECTORY = os.environ.get(
    "IRSX_INDEX_DIRECTORY", os.path.join(IRSX_CACHE_DIRECTORY, "CSV"))

IRS_INDEX_BASE = "https://apps.irs.gov/pub/epostcard/990/xml/%s/index_%s.csv"

KNOWN_SCHEDULES = [
                "IRS990", "IRS990EZ", "IRS990PF", "IRS990ScheduleA",
                "IRS990ScheduleB", "IRS990ScheduleC", "IRS990ScheduleD",
                "IRS990ScheduleE", "IRS990ScheduleF", "IRS990ScheduleG",
                "IRS990ScheduleH", "IRS990ScheduleI", "IRS990ScheduleJ",
                "IRS990ScheduleK", "IRS990ScheduleL", "IRS990ScheduleM",
                "IRS990ScheduleN", "IRS990ScheduleO", "IRS990ScheduleR",
                "ReturnHeader990x"
]

# these could get pushed to metadata directory? 

ALLOWED_VERSIONSTRINGS = [
    '2013v3.0', '2013v3.1', '2013v4.0', '2014v5.0', '2014v6.0',
    '2015v2.0', '2015v2.1', '2015v3.0', '2016v3.0', '2016v3.1',
    '2017v2.0', '2017v2.1', '2017v2.2', '2017v2.3', '2018v3.0',
    '2018v3.1', '2018v3.2', '2018v3.3', '2019v5.0', '2019v5.1', 
    '2019v5.2', '2020v1.0', '2020v2.0', '2020v3.0', '2020v4.0',
    '2020v4.1', '2020v4.2', '2020v1.1'
]

# 2020 is experimental
# see https://www.irs.gov/charities-non-profits/ty2020-xml-schemas-and-business-rules-for-exempt-organizations-modernized-e-file

# We can capture the group structure for these so it doesn't break
# but these versions ARE NOT supported and aren't mapped to IRSx variables
CSV_ALLOWED_VERSIONSTRINGS = ALLOWED_VERSIONSTRINGS + [
    '2010v3.2', '2010v3.4', '2010v3.6', '2010v3.7', '2011v1.2', '2011v1.3',
    '2011v1.4', '2011v1.5', '2012v2.0', '2012v2.1', '2012v2.2', '2012v2.3', 
    '2012v3.0'
]

METADATA_DIRECTORY = (os.path.join(IRS_READER_ROOT, "metadata"))

KEYERROR_LOG = os.path.join(IRS_READER_ROOT, "keyerrors.log")
LOG_KEY = 'xml'

mkdir_p([WORKING_DIRECTORY, INDEX_DIRECTORY])

try:
    from .local_settings import *
except ImportError:
    pass



import os
import sys
import io
import xmltodict
import json
from collections import OrderedDict
from xml.parsers.expat import ExpatError
from .type_utils import dictType, orderedDictType, listType, \
    unicodeType, noneType, strType

from .file_utils import stream_download, validate_object_id, \
    get_local_path

from .settings import KNOWN_SCHEDULES, IRS_READER_ROOT


class InvalidXMLException(Exception):
    pass


class FileMissingException(Exception):
    pass

class Filing(object):

    def __init__(self, object_id, filepath=None, URL=None, json=None):
        """ Filepath is the location of the file locally;
            URL is it's remote location (if not default)
            Ignore these and defaults will be used.
            If filepath is set, URL is ignored.
            json is a json representation of the data, so if given, 
            no file will be downloaded.
        """
        self.raw_irs_dict = None        # The parsed xml will go here
        self.version_string = None      # Version number here

        self.object_id = validate_object_id(object_id)
        self.result = None
        self.processed = False
        self.keyerrors = None
        self.csv_result = None

        if json:
            self.json = json
            self.input_type = 'json'
        else:
            self.json = None
            self.input_type = 'xml'
            if filepath:
                self.filepath = filepath
            else:
                self.filepath = get_local_path(self.object_id)

                if URL:
                    self.URL = URL
                

    def _download(self, force_overwrite=False, verbose=False):
        """
        Files are no longer downloadable. 
        """
        
        if os.path.isfile(self.filepath):
            return True
        else: 
            raise FileMissingException(
                "Filing not available, try downloading with irsx_retrieve [ YEAR ]"
                )
        

    def _denamespacify(self,entity):
        """
        It's legal to include namespaces in the xml tags, e.g. irs:Return instead of Return
        This is very rare; see 201940149349301304_public.xml for an example.
        """
        thisentitytype = type(entity)
        if thisentitytype == orderedDictType:
            newOD = OrderedDict()
            for key in entity.keys():
                newkey = key
                if ":" in key:
                    newkey = key.split(":")[1]
                newvalue = entity[key]
                if type(newvalue) == listType or type(newvalue) == orderedDictType:
                    newvalue = self._denamespacify(newvalue)
                newOD[newkey] = newvalue
            return newOD

        elif thisentitytype == listType:
            newlist = list()
            for item in entity:
                newvalue = item
                if type(newvalue) == listType or type(newvalue) == orderedDictType:
                    newvalue = self._denamespacify(newvalue)
                newlist.append(newvalue)
            return newlist
        else: 
            return entity


    def _set_dict_from_xml(self):
        # io works across python2 and 3, and allows an encoding arg        
        with io.open(self.filepath, 'r', encoding='utf-8-sig') as fh:
            raw_file = fh.read()
            try:

                self.raw_irs_dict = self._denamespacify(xmltodict.parse(raw_file))
            except ExpatError:
                raise InvalidXMLException(
                    "\nXML Parse error in " + self.filepath \
                    + "\nFile may be damaged or incomplete.\n"\
                    + "Try erasing this file and downloading again."
                )
            try:
                self.raw_irs_dict['Return']
            except KeyError:
                raise InvalidXMLException(
                    "'Return' element not located in" + self.filepath \
                    + "\nFile may be damaged or incomplete.\n" \
                    + "Try erasing this file and downloading again."
                )



    def _set_dict_from_json(self):

        self.raw_irs_dict = self.json

    def _set_version(self):
        self.version_string = self.raw_irs_dict['Return']['@returnVersion']

    def _set_ein(self):
        self.ein = self.raw_irs_dict['Return']['ReturnHeader']['Filer']['EIN']

    def _set_schedules(self):
        """ Attach the known and unknown schedules """
        self.schedules = ['ReturnHeader990x', ]
        self.otherforms = []
        for sked in self.raw_irs_dict['Return']['ReturnData'].keys():
            if not sked.startswith("@"):
                if sked in KNOWN_SCHEDULES:
                    self.schedules.append(sked)
                else:
                    self.otherforms.append(sked)

    def get_object_id(self):
        return self.object_id

    def get_schedule(self, skedname):
        if skedname == 'ReturnHeader990x':
            return self.raw_irs_dict['Return']['ReturnHeader']
        elif skedname in self.schedules:
            return self.raw_irs_dict['Return']['ReturnData'][skedname]
        else:
            return None

    def get_ein(self):
        return self.ein

    def get_otherform(self, skedname):
        if skedname in self.otherforms:
            return self.raw_irs_dict['Return']['ReturnData'][skedname]
        else:
            return None

    def get_filepath(self):
        return self.filepath

    def get_version(self):
        return self.version_string

    def get_raw_irs_dict(self):
        return self.raw_irs_dict

    def list_schedules(self):
        return self.schedules

    def set_result(self, result):
        self.result = result

    def get_result(self):
        return self.result

    def set_csv_result(self, csv_result):
        self.csv_result = csv_result
        
    def get_csv_result(self):
        return self.csv_result

    def set_keyerrors(self, keyerrorlist):
        self.keyerrors = keyerrorlist

    def get_keyerrors(self):
        return self.keyerrors
        
    def get_unparsed_json(self):
        """ Json dicts are unordered """ 
        return json.dumps(self.raw_irs_dict)

    def get_type(self):
        if 'IRS990' in self.schedules:
            return 'IRS990'
        elif 'IRS990EZ' in self.schedules:
            return 'IRS990EZ'
        elif 'IRS990PF' in self.schedules:
            return 'IRS990PF'    
        else:
            raise Exception("Missing 990/990EZ/990PF-is this filing valid?")

    def get_parsed_sked(self, skedname):
        """ Returns an array because multiple sked K's are allowed"""
        if not self.processed:
            raise Exception("Filing must be processed to return parsed sked")
        if skedname in self.schedules:
            matching_skeds = []
            for sked in self.result:
                if sked['schedule_name']==skedname:
                    matching_skeds.append(sked)
            return matching_skeds
        else:
            return []

    def process(self, verbose=False):
        # don't reprocess inadvertently
        if not self.processed:
            self.processed=True
            if self.json:
                self._set_dict_from_json()
            else:
                
                self._download(verbose=verbose)
                self._set_dict_from_xml()

            self._set_version()
            self._set_ein()
            self._set_schedules()



import os
import sys
import collections
#import logging
from datetime import datetime
from .settings import METADATA_DIRECTORY, KEYERROR_LOG
from .sked_dict_reader import SkedDictReader
from .type_utils import listType
if sys.version_info >= (3, 0):
    import csv
else:
    import unicodecsv as csv


class Standardizer(object):
    """
    This reads metadata .csv files, which it uses to standardize
    ordered dicts. For documentation, see Documentizer below. 
    """

    def __init__(self):
        #self.show_documentation = documentation
        self.groups = {}
        self.variables = {}
        self.schedule_parts = {}

        # This is overridden for Documentizer class below
        self.variable_columns =['db_table', 'db_name']

        self._make_groups()
        self._make_variables()


    def _make_groups(self):
        group_filepath = os.path.join(METADATA_DIRECTORY, 'groups.csv')
        with open(group_filepath, 'r') as reader_fh:
            reader = csv.DictReader(reader_fh)
            for row in reader:
                self.groups[row['xpath']] = row
        return True

    def _make_variables(self):
        variable_filepath = os.path.join(METADATA_DIRECTORY, 'variables.csv')
        with open(variable_filepath, 'r') as variable_fh:
            reader = csv.DictReader(variable_fh)
            for row in reader:
                vardict = {}
                for col in self.variable_columns:
                    vardict[col]=row[col]
                self.variables[row['xpath']] = vardict


        return True

    def get_groups(self):
        return self.groups

    def get_var(self, var_xpath, version=None):
        if version:
            raise Exception("Version checking is not implemented")
        return (self.variables[var_xpath])

    def get_documentation_status(self):
        return False


class Documentizer(Standardizer):
    """ Like Standardizer, but returns canonical documentation info from 2016 version """

    def __init__(self, versions=False):
        self.groups = {}
        self.variables = {}
        self.schedule_parts = {}

        self.variable_columns =[
            'db_table', 'db_name', 'ordering', 
            'line_number', 'description', 'db_type',
            'irs_type', 'xpath'
        ]
        if versions:
            self.variable_columns = self.variable_columns + ['version_start', 'version_end']

        self._make_schedule_parts()
        self._make_groups()
        self._make_variables()

    def get_documentation_status(self):
        return True

    def _make_schedule_parts(self):
        part_filepath = os.path.join(METADATA_DIRECTORY, 'schedule_parts.csv')
        with open(part_filepath, 'r') as reader_fh:
            reader = csv.DictReader(reader_fh)
            for row in reader:
                self.schedule_parts[row['parent_sked_part']] = {
                    'name': row['part_name'],
                    'ordering': row['ordering'],
                    'parent_sked': row['parent_sked'],
                    'parent_sked_part': row['parent_sked_part'],
                    'is_shell': row['is_shell']

                }
        return True

    def get_schedule_parts(self):
        return self.schedule_parts

    def part_ordering(self, partname):
        try:
            result = int(self.schedule_parts[partname]['ordering'])
            return result
        except KeyError:
            return None

    def group_ordering(self, groupname):
        try:
            return self.groups[groupname]['ordering']
        except KeyError:
            return None

    def get_groups_by_sked(self, sked):
        groups = []
        for thisgroup in self.groups.keys():
            if self.groups[thisgroup]['parent_sked'] == sked:
                groups.append(self.groups[thisgroup])
        return groups

    def get_parts_by_sked(self, sked):
        parts = []
        for thispart in self.schedule_parts.keys():
            #print(self.schedule_parts[thispart])
            if self.schedule_parts[thispart]['parent_sked'] == sked:
                parts.append(self.schedule_parts[thispart])
        return parts

    def get_variables(self):
        return self.variables



class VersionDocumentizer(object):
    """
    Returns version-specific line number and documentation.
    """

    def __init__(self):
        self.line_numbers = {}
        self.descriptions = {}

        self._make_line_numbers()
        self._make_descriptions()

    def check_version(self, versionstring, start_year, end_year):
        versionyear = int(versionstring.split("v")[0])
        valid_start = versionyear >= int(start_year)
        valid_end = not end_year or versionyear <= int(end_year)
        result = valid_start and valid_end
        return result

    def _make_line_numbers(self):
        filepath = os.path.join(METADATA_DIRECTORY, 'line_numbers.csv')
        with open(filepath, 'r') as reader_fh:
            reader = csv.DictReader(reader_fh)

            for row in reader:
                try:
                    self.line_numbers[row['xpath']]
                    self.line_numbers[row['xpath']].append(row)

                except KeyError:
                    self.line_numbers[row['xpath']] = [row]

    def _make_descriptions(self):
        filepath = os.path.join(METADATA_DIRECTORY, 'descriptions.csv')
        with open(filepath, 'r') as reader_fh:
            reader = csv.DictReader(reader_fh)

            for row in reader:
                try:
                    self.descriptions[row['xpath']]
                    self.descriptions[row['xpath']].append(row)

                except KeyError:
                    self.descriptions[row['xpath']] = [row]


    def get_line_number(self, xpath, version_string):
        candidate_rows = []
        try:
            candidate_rows = self.line_numbers[xpath]
        except KeyError:
            return None

        for row in candidate_rows:
            if self.check_version(version_string, row['version_start'], row['version_end']):
                return row['line_number']

        return None

    def get_description(self, xpath, version_string):
        candidate_rows = []
        try:
            candidate_rows = self.descriptions[xpath]
        except KeyError:
            return None
        for row in candidate_rows:
            if self.check_version(version_string, row['version_start'], row['version_end']):
                return row['description']
        return None




import sys
import os
import argparse
from zipfile import ZipFile
from .file_utils import stream_download
from .settings import WORKING_DIRECTORY

IRS_location = "https://apps.irs.gov/pub/epostcard/990/xml/%s/download990xml_%s"

ref_url = "https://www.irs.gov/charities-non-profits/form-990-series-downloads"
# How many files are available per year? 
# https://www.irs.gov/charities-non-profits/form-990-series-downloads
number_of_files = {
    '2022':0,
    '2021':6,
    '2020':8,
    '2019':8,
    '2018':7,
    '2017':7,
    '2016':6,
    '2015':2
}

def get_cli_retrieve_parser():
    parser = argparse.ArgumentParser("Irsreader")
    parser.add_argument(
        "year",
        nargs='+',
        help='4-digit year to retrieve, '
    )

    parser.add_argument(
        '--verbose',
        dest='verbose',
        action='store_const',
        const=True, default=False,
        help='Verbose output'
    )
    return parser


def download_unzip_erase(remote_url, verbose=False):
    local_name = remote_url.split("/")[-1]
    local_path = os.path.join(WORKING_DIRECTORY, local_name)

    if verbose:
        print("Downloading %s to %s" % (remote_url, local_path))
    stream_download(remote_url, local_path, verbose=verbose)

    with ZipFile(local_path, 'r') as zipObj:
    # Extract all the contents of zip file in different directory
        print('Unzipping %s to %s' % (local_path, WORKING_DIRECTORY))
        zipObj.extractall(WORKING_DIRECTORY)

    print("Cleaning up, removing raw file.")
    os.remove(local_path)

def unload_zipfile_by_year(year, verbose=False):
    print("Retrieving zipfiles for year %s" % year)
    if verbose:
        print("Running verbose")

    num_files = number_of_files[year]
    location_base = IRS_location % (year, year)
    file_list = []

    if num_files == 0:
        file_list.append(location_base + ".zip")

    if num_files > 0:
        for i in range(1, num_files+1):
            file_list.append(location_base + "_" + str(i) + ".zip")

    for this_file in file_list:
        download_unzip_erase(this_file, verbose=verbose)


def run_cli_retrieve_main(args_read):
    print("""
    Please visit https://www.irs.gov/charities-non-profits/form-990-series-downloads
    To see if any additional files are available. 
    """)
    for year in args_read.year:
        print("Processing %s files for year %s" % (year, number_of_files[year]))
        unload_zipfile_by_year(year, verbose=args_read.verbose)

def main(args=None):
    parser = get_cli_retrieve_parser()
    args = parser.parse_args()
    run_cli_retrieve_main(args)

if __name__ == "__main__":
    main()


import errno
import os


def mkdir_p(paths):
    """ Makedirs, from http://stackoverflow.com/a/600612 """
    for path in paths:
        try:
            os.makedirs(path)
        except OSError as exc:  # Python >2.5
            if exc.errno == errno.EEXIST and os.path.isdir(path):
                pass
            else:
                raise



import argparse

from .filing import Filing
from .settings import KNOWN_SCHEDULES, IRS_READER_ROOT
from .xmlrunner import XMLRunner
from .text_format_utils import *


def get_parser():
    parser = argparse.ArgumentParser("irsx")

    parser.add_argument(
        'object_ids',
        metavar='object_ids',
        type=int,
        nargs='+',
        help='object ids'
    )
    parser.add_argument(
        '--verbose',
        dest='verbose',
        action='store_const',
        const=True, default=False,
        help='Verbose output'
    )
    parser.add_argument(
        "--schedule",
        choices=KNOWN_SCHEDULES,
        default=None,
        help='Get only that schedule'
    )
    parser.add_argument(
        "--xpath",
        dest='documentation',
        action='store_const',
        const=True, default=False,
        help='show xpath in text format'
    )
    parser.add_argument(
        "--format",
        choices=['json', 'csv', 'txt'],
        default='json',
        help='Output format'
    )
    parser.add_argument(
        "--file",
        default=None,
        help='Write result to file'
    )
    parser.add_argument(
        '--list_schedules',
        dest='list_schedules',
        action='store_const',
        const=True,
        default=False,
        help='Only list schedules'
    )
    return parser


def run_main(args_read):

    csv_format = args_read.format == 'csv' or args_read.format == 'txt'
    xml_runner = XMLRunner(
        documentation=args_read.documentation,
        csv_format=csv_format
    )

    # Use the standardizer that was init'ed by XMLRunner
    standardizer = xml_runner.get_standardizer()

    for object_id in args_read.object_ids:
        if args_read.verbose:
            print("Processing filing %s" % object_id)
            if args_read.file:
                print("Printing result to file %s" % args_read.file)

        if args_read.list_schedules:
            this_filing = Filing(object_id)
            this_filing.process()
            print(this_filing.list_schedules())
            return True  # we're done, ignore any other commands

        else:
            if args_read.schedule:
                parsed_filing = xml_runner.run_sked(
                    object_id,
                    args_read.schedule,
                    verbose=args_read.verbose
                )
            else:
                parsed_filing = xml_runner.run_filing(
                    object_id,
                    verbose=args_read.verbose
                )

        if args_read.format == 'json':
            to_json(parsed_filing.get_result(), outfilepath=args_read.file)

        elif args_read.format == 'csv':
                to_csv(
                    parsed_filing,
                    object_id=object_id,
                    standardizer=standardizer,
                    documentation=args_read.documentation,
                    outfilepath=args_read.file
                )

        elif args_read.format == 'txt':
                to_txt(
                    parsed_filing,
                    standardizer=standardizer,
                    documentation=args_read.documentation,
                    outfilepath=args_read.file
                )


def main(args=None):
    parser = get_parser()
    args_read = parser.parse_args()
    run_main(args_read)
    print("\n")

if __name__ == "__main__":
    main()



import collections
#  Mostly from: http://stackoverflow.com/a/6027615


def flatten(d, parent_key='', sep='/'):
    items = []
    if d:
        for k, v in d.items():
            new_key = parent_key + sep + k if parent_key else k
            if isinstance(v, collections.abc.MutableMapping):
                items.extend(flatten(v, new_key, sep=sep).items())
            else:
                new_key = new_key.replace("/#text","")
                items.append((new_key, v))
        return dict(items)
    else:
        return {}




ignorable_keyerrors = ['/ReturnHeader/BuildTS']

## Todo: put in 2013 / 2015 series canonicals.
# 2013 vars that no longer exist
discontinued_2013_vars = [ '/IRS990ScheduleA/CertificationInd', '/IRS990ScheduleA/Contribution35ControlledInd', '/IRS990ScheduleA/ContributionControllerInd', '/IRS990ScheduleA/ContributionFamilyInd', '/IRS990ScheduleA/Form990ScheduleAPartIVGrp/ExplanationTxt', '/IRS990ScheduleA/SupportedOrgInformationGrp/SupportedOrgNotifiedInd', '/IRS990ScheduleA/SupportedOrgInformationGrp/USOrganizedInd', '/IRS990ScheduleH/HospitalFcltyPoliciesPrctcGrp/AdoptBudgetInd', '/IRS990ScheduleH/HospitalFcltyPoliciesPrctcGrp/AdoptImplementationStrategyInd', '/IRS990ScheduleH/HospitalFcltyPoliciesPrctcGrp/AllNeedsAddressedInd', '/IRS990ScheduleH/HospitalFcltyPoliciesPrctcGrp/AttachedToInvoiceInd', '/IRS990ScheduleH/HospitalFcltyPoliciesPrctcGrp/AvailableOnRequestInd', '/IRS990ScheduleH/HospitalFcltyPoliciesPrctcGrp/BodyAttachmentsInd', '/IRS990ScheduleH/HospitalFcltyPoliciesPrctcGrp/DevelopCommunityWidePlanInd', '/IRS990ScheduleH/HospitalFcltyPoliciesPrctcGrp/ExecCommunityWidePlanInd', '/IRS990ScheduleH/HospitalFcltyPoliciesPrctcGrp/ExecImplementationStrategyInd', '/IRS990ScheduleH/HospitalFcltyPoliciesPrctcGrp/FPGUsedDeterEligFreeCareInd', '/IRS990ScheduleH/HospitalFcltyPoliciesPrctcGrp/FPGUsedDetermEligDscntCareInd', '/IRS990ScheduleH/HospitalFcltyPoliciesPrctcGrp/IncludeOperationalPlanInd', '/IRS990ScheduleH/HospitalFcltyPoliciesPrctcGrp/LawsuitInd', '/IRS990ScheduleH/HospitalFcltyPoliciesPrctcGrp/LiensOnResidencesInd', '/IRS990ScheduleH/HospitalFcltyPoliciesPrctcGrp/MedicaidMedicareInd', '/IRS990ScheduleH/HospitalFcltyPoliciesPrctcGrp/OtherNeedsAddressedInd', '/IRS990ScheduleH/HospitalFcltyPoliciesPrctcGrp/PermitBodyAttachmentsInd', '/IRS990ScheduleH/HospitalFcltyPoliciesPrctcGrp/PermitLawsuitInd', '/IRS990ScheduleH/HospitalFcltyPoliciesPrctcGrp/PermitLienOnResidenceInd', '/IRS990ScheduleH/HospitalFcltyPoliciesPrctcGrp/PostedInAdmissionOfficeInd', '/IRS990ScheduleH/HospitalFcltyPoliciesPrctcGrp/PostedInEmergencyRoomInd', '/IRS990ScheduleH/HospitalFcltyPoliciesPrctcGrp/PrioritizeHealthNeedsInd', '/IRS990ScheduleH/HospitalFcltyPoliciesPrctcGrp/PrioritizeServicesInd', '/IRS990ScheduleH/HospitalFcltyPoliciesPrctcGrp/ProvidedOnAdmissionInd', '/IRS990ScheduleH/HospitalFcltyPoliciesPrctcGrp/StateRegulationInd', '/IRS990ScheduleH/HospitalFcltyPoliciesPrctcGrp/UninsuredDiscountInd']
# 2015 skedh vars removed
discontinued_2015_vars = ['/IRS990ScheduleH/HospitalFcltyPoliciesPrctcGrp/AverageNegotiatedRatesInd', '/IRS990ScheduleH/HospitalFcltyPoliciesPrctcGrp/DocumentedEligDeterminationInd', '/IRS990ScheduleH/HospitalFcltyPoliciesPrctcGrp/FAPNoticeDisplayedInd', '/IRS990ScheduleH/HospitalFcltyPoliciesPrctcGrp/FAPNotifiedAllPatientsInd', '/IRS990ScheduleH/HospitalFcltyPoliciesPrctcGrp/FAPNotifiedBeforeDischargeInd', '/IRS990ScheduleH/HospitalFcltyPoliciesPrctcGrp/FAPNotifiedUponAdmissionInd', '/IRS990ScheduleH/HospitalFcltyPoliciesPrctcGrp/InformationGapsInd', '/IRS990ScheduleH/HospitalFcltyPoliciesPrctcGrp/LowestNegotiatedRatesInd', '/IRS990ScheduleH/HospitalFcltyPoliciesPrctcGrp/MedicareRatesInd', '/IRS990ScheduleH/HospitalFcltyPoliciesPrctcGrp/OtherMethodUsedInd']

ignorable = {}
for key in ignorable_keyerrors + discontinued_2013_vars + discontinued_2015_vars:
    ignorable[key] = 1

def ignorable_keyerror(xpath):
    try:
        ignorable[xpath]
        return True
    except KeyError:
        return False



import os
from .dir_utils import mkdir_p


IRS_READER_ROOT = "/path/to/irsreader/990-xml-reader"

# This is the URL to amazon's bucket, could use another synced to it
IRS_XML_HTTP_BASE = "https://s3.amazonaws.com/irs-form-990"

# The directory we put files in while we're processing them
WORKING_DIRECTORY = (os.path.join(IRS_READER_ROOT, "XML") )

# Helpful to keep these around for lookup purposes
INDEX_DIRECTORY = (os.path.join(IRS_READER_ROOT, "CSV") )

mkdir_p([WORKING_DIRECTORY, INDEX_DIRECTORY])


import sys
import argparse
from datetime import date
from .file_utils import get_index_file_URL, get_local_index_path, \
    stream_download

this_year = date.today().year
INDEXED_YEARS = [str(i) for i in range(2011, this_year+1)]


def get_cli_index_parser():
    parser = argparse.ArgumentParser("Irsreader")
    parser.add_argument(
        "--year",
        choices=INDEXED_YEARS,
        default=None,
        help='Optionally update an index file'
    )

    parser.add_argument(
        '--verbose',
        dest='verbose',
        action='store_const',
        const=True, default=False,
        help='Verbose output'
    )
    return parser


def get_indexfile_by_year(year, verbose=False):
    localpath = get_local_index_path(year)
    remoteurl = get_index_file_URL(year)
    if verbose:
        print(
            "Getting index file for year: %s remote=%s local=%s"
            % (year, remoteurl, localpath)
        )
    stream_download(remoteurl, localpath, verbose=verbose)


def run_cli_index_main(args_read):
    if args_read.year:
        get_indexfile_by_year(args_read.year, verbose=args_read.verbose)
    else:
        for year in INDEXED_YEARS:
            get_indexfile_by_year(year, verbose=args_read.verbose)


def main(args=None):
    parser = get_cli_index_parser()
    args = parser.parse_args()
    run_cli_index_main(args)


if __name__ == "__main__":
    main()



import logging
from .settings import LOG_KEY, KEYERROR_LOG


def configure_logging(name=LOG_KEY):
    logger = logging.getLogger(name)
    logger.setLevel(logging.INFO)
    # Format
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s")

    # Setup console logging
    ch = logging.StreamHandler()
    ch.setLevel(logging.WARNING)
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    # Setup file logging
    fh = logging.FileHandler(KEYERROR_LOG)
    fh.setLevel(logging.INFO)
    fh.setFormatter(formatter)
    logger.addHandler(fh)
    return logger



from collections import OrderedDict

# Type names for comparison
# Checking types is a bad idea, blahdy blahdy blahdy

dictType = type(dict())
orderedDictType = type(OrderedDict())
listType = type(list())
unicodeType = type(u'')
noneType = type(None)
strType = type('')



import json
import sys
import codecs
import re
import csv
import unicodecsv
 
from .standardizer import Standardizer, Documentizer, VersionDocumentizer


BRACKET_RE = re.compile(r'\[.*?\]')

ASTERISKS = "****************"

def debracket(string):
    """ Eliminate the bracketed var names in doc, line strings """
    result = re.sub(BRACKET_RE, ';', str(string))
    result = result.lstrip(';')
    result = result.lstrip(' ')
    result = result.replace('; ;',';')
    return result

def most_recent(semicolon_delimited_string):
    result = semicolon_delimited_string.split(";")[-1]
    return result

def to_json(data, outfilepath=None):
    if data:
        if outfilepath:
            with open(outfilepath, 'w') as outfile:
                json.dump(data, outfile)
        else:
            if hasattr(sys.stdout, "buffer"):
                sys.stdout = codecs.getwriter("utf-8")(sys.stdout.buffer, "strict")
                json.dump(data, sys.stdout)
            else:
                json.dump(data, sys.stdout)

def to_csv(parsed_filing, object_id=None, standardizer=None, documentation=True, vd=None, outfilepath=None):
    if not vd:
        vd = VersionDocumentizer()
    stdout = getattr(sys.stdout, 'buffer', sys.stdout)
    if outfilepath:
        stdout = open(outfilepath, 'wb')  # or 'wb' ?

    fieldnames = []
    fieldnames = [ 
            'object_id', 'form', 'line_number', 'description', 'value', 'variable_name',
            'xpath', 'in_group', 'group_name', 'group_index'
        ]
    writer = unicodecsv.DictWriter(
        stdout,
        fieldnames=fieldnames,
        encoding='utf-8',
        quoting=csv.QUOTE_MINIMAL
    )
    writer.writeheader()   # this fails in python3? 
    results = parsed_filing.get_result()

    if results:
        for result in results:
            for this_result in result['csv_line_array']:

                vardata = None
                try:
                    vardata = standardizer.get_var(this_result['xpath'])
                except KeyError:
                    pass
                if vardata:
                    this_result['variable_name'] = vardata['db_table'] + "." + vardata['db_name']

                raw_line_num = vd.get_line_number(
                    this_result['xpath'], 
                    parsed_filing.get_version()
                )
                this_result['line_number'] =  debracket(raw_line_num)

                raw_description = vd.get_description(
                    this_result['xpath'], 
                    parsed_filing.get_version()
                )
                this_result['description'] =  debracket(raw_description)
                this_result['form'] = this_result['xpath'].split("/")[1]
                this_result['object_id'] = object_id
                writer.writerow(this_result)


def to_txt(parsed_filing, standardizer=None, documentation=True, vd=None, outfilepath=None):
    if not vd:
        vd = VersionDocumentizer()
    results = parsed_filing.get_result()
    this_sked_name = None
    if outfilepath:
        outfile = open(outfilepath, 'w')
    if results:
        for result in results:
            for this_result in result['csv_line_array']:

                #### Collect the variables we need
                vardata = None
                textoutput = "\n"     # This is what we'll eventually write out
                this_result['form'] = this_result['xpath'].split("/")[1]
                try:
                    vardata = standardizer.get_var(this_result['xpath'])
                except KeyError:
                    pass
                if vardata:
                    this_result['variable_name'] = vardata['db_table'] + "." + vardata['db_name']

                raw_line_num = vd.get_line_number(
                    this_result['xpath'], 
                    parsed_filing.get_version()
                )
                this_result['line_number'] =  debracket(raw_line_num)

                raw_description = vd.get_description(
                    this_result['xpath'], 
                    parsed_filing.get_version()
                )
                this_result['description'] =  debracket(raw_description)

                #### Write the output, now that we've got the vars 

                if this_sked_name != this_result['form']:
                    textoutput += "\n\n\n" + ASTERISKS + "\tSchedule %s\n" % this_result['form']
                    this_sked_name = this_result['form']
                
                textoutput += "\n" + ASTERISKS + "\n  Value: '%s'\nForm: %s\nLine:%s\nDescription:%s" % (
                    this_result['value'], 
                    this_result['form'],
                    this_result['line_number'],
                    this_result['description'], 
                )

                if documentation:
                    textoutput += "\nXpath:%s" % (this_result['xpath'])

                if this_result['in_group']:
                    textoutput += "\nGroup: %s group_index %s" % (
                        this_result['group_name'], 
                        this_result['group_index']
                    )
                else:
                    textoutput += "\nGroup:"
                    
                if outfilepath:
                    outfile.write(textoutput)
                else:
                    sys.stdout.write(textoutput)

    if outfilepath:
        outfile.close()


import re
import os
import requests

from datetime import datetime
from .settings import IRS_XML_HTTP_BASE, WORKING_DIRECTORY, INDEX_DIRECTORY, IRS_INDEX_BASE

OBJECT_ID_RE = re.compile(r'20\d{16}')

# Not sure how much detail we need to go into here
OBJECT_ID_MSG = """
This appears not to be an IRS object id.
The ID should be 18 digits long and start with
the four digit year, e.g. 201642229349300909

To find the object id, see the yearly index csv files.
"""


def stream_download(url, target_path, verbose=False):
    """ Download a large file without loading it into memory. """
    response = requests.get(url, stream=True)
    handle = open(target_path, "wb")
    if verbose:
        print("Beginning streaming download of %s" % url)
        start = datetime.now()
        try:
            content_length = int(response.headers['Content-Length'])
            content_MB = content_length/1048576.0
            print("Total file size: %.2f MB" % content_MB)
        except KeyError:
            pass      # allow Content-Length to be missing
    for chunk in response.iter_content(chunk_size=512):
        if chunk:     # filter out keep-alive new chunks
            handle.write(chunk)

    if verbose:
        print(
            "Download completed to %s in %s" %
            (target_path, datetime.now() - start))


def validate_object_id(object_id):
    """ It's easy to make a mistake entering these, validate the format """
    result = re.match(OBJECT_ID_RE, str(object_id))
    if not result:
        print("'%s' appears not to be a valid 990 object_id" % object_id)
        raise RuntimeError(OBJECT_ID_MSG)
    return object_id


# Files are no longer available on S3
# def get_s3_URL(object_id):
#     return ("%s/%s_public.xml" % (IRS_XML_HTTP_BASE, object_id))


def get_local_path(object_id):
    file_name = "%s_public.xml" % object_id
    return os.path.join(WORKING_DIRECTORY, file_name)


def get_index_file_URL(year):
    index_file = IRS_INDEX_BASE % (year, year)
    print("index file %s %s" % (year, index_file))
    return index_file


def get_local_index_path(year):
    csv_file_name = "index_%s.csv" % year
    return os.path.join(INDEX_DIRECTORY, csv_file_name)



from .filing import Filing
from .standardizer import Standardizer, Documentizer, VersionDocumentizer
from .sked_dict_reader import SkedDictReader
# from .log_utils import configure_logging
from .type_utils import listType

from .settings import WORKING_DIRECTORY, ALLOWED_VERSIONSTRINGS, CSV_ALLOWED_VERSIONSTRINGS

class XMLRunner(object):
    """ Load a Standardizer just once while running multiple filings
        Return Filing objects with results, keyerrors set
    """
    def __init__(self, documentation=False, standardizer=None, csv_format=False):
        self.documentation = documentation
        self.csv_format = csv_format

        if documentation:
            if not standardizer:    
                self.standardizer = Documentizer()
        else:
            if standardizer:
                self.standardizer = standardizer
            else:
                self.standardizer = Standardizer()
        self.group_dicts = self.standardizer.get_groups()
        self.whole_filing_data = []
        self.filing_keyerr_data = []

    def get_standardizer(self):
        return self.standardizer

    def _run_schedule_k(self, sked, object_id, sked_dict, path_root, ein):
        assert sked == 'IRS990ScheduleK'
        if type(sked_dict) == listType:
            for individual_sked in sked_dict:
                doc_id = individual_sked['@documentId']
                reader = SkedDictReader(
                    self.standardizer,
                    self.group_dicts,
                    object_id,
                    ein,
                    documentId=doc_id,
                    documentation=self.documentation,
                    csv_format=self.csv_format,
                )

                result = reader.parse(individual_sked, parent_path=path_root)
                self.whole_filing_data.append({
                    'schedule_name': sked,
                    'groups': result['groups'],
                    'schedule_parts': result['schedule_parts'],
                    'csv_line_array':result['csv_line_array']

                })
        else:
            reader = SkedDictReader(
                self.standardizer,
                self.group_dicts,
                object_id,
                ein,
                documentation=self.documentation,
                csv_format=self.csv_format,
            )

            result = reader.parse(sked_dict, parent_path=path_root)
            self.whole_filing_data.append({
                'schedule_name': sked,
                'groups': result['groups'],
                'schedule_parts': result['schedule_parts'],
                'csv_line_array':result['csv_line_array']
            })

    def _run_schedule(self, sked, object_id, sked_dict, ein):
        path_root = "/" + sked
        # Only sked K (bonds) is allowed to repeat
        if sked == 'IRS990ScheduleK':
            self._run_schedule_k(sked, object_id, sked_dict, path_root, ein)

        else:
            reader = SkedDictReader(
                self.standardizer,
                self.group_dicts,
                object_id,
                ein,
                documentation=self.documentation,
                csv_format=self.csv_format,
            )
            if sked == 'ReturnHeader990x':
                path_root = "/ReturnHeader"
            result = reader.parse(sked_dict, parent_path=path_root)

            self.whole_filing_data.append({
                'schedule_name': sked,
                'groups': result['groups'],
                'schedule_parts': result['schedule_parts'],
                'csv_line_array':result['csv_line_array']

            })

            if len(result['group_keyerrors']) > 0 or len(result['keyerrors'])> 0:
                self.filing_keyerr_data.append({
                    'schedule_name': sked,
                    'group_keyerrors':result['group_keyerrors'],
                    'keyerrors':result['keyerrors'],
                })

    def run_filing(self, object_id, verbose=False):
        self.whole_filing_data = []
        self.filing_keyerr_data = []
        this_filing = Filing(object_id)
        this_filing.process(verbose=verbose)
        this_version = this_filing.get_version()
        if verbose:
            print("Filing %s is version %s" % (object_id, this_version))
        if this_version in ALLOWED_VERSIONSTRINGS or ( self.csv_format and this_version in CSV_ALLOWED_VERSIONSTRINGS ):
            this_version = this_filing.get_version()
            schedules = this_filing.list_schedules()
            ein = this_filing.get_ein()
            self.whole_filing_data = []
            for sked in schedules:
                sked_dict = this_filing.get_schedule(sked)
                self._run_schedule(sked, object_id, sked_dict, ein)

            this_filing.set_result(self.whole_filing_data)
            this_filing.set_keyerrors(self.filing_keyerr_data)
            if verbose and not self.csv_format:   # csv format works on years with many, many keyerrors, 
                if len(self.filing_keyerr_data)>0:
                    print("In %s keyerrors: %s" % (object_id, self.filing_keyerr_data))
                else:
                    print("No keyerrors found")
            return this_filing
        else:
            print("Filing version %s isn't supported for this operation" % this_version )
            return this_filing

    """
    def run_from_filing_obj(self, this_filing, verbose=False):  
        
        #Run from a pre-created filing object.
        
        self.whole_filing_data = []
        self.filing_keyerr_data = []
        this_filing.process(verbose=verbose)
        object_id = this_filing.get_object_id()
        this_version = this_filing.get_version()
        if this_version in ALLOWED_VERSIONSTRINGS:
            this_version = this_filing.get_version()
            schedules = this_filing.list_schedules()
            ein = this_filing.get_ein()
            for sked in schedules:
                sked_dict = this_filing.get_schedule(sked)
                self._run_schedule(sked, object_id, sked_dict, ein)
            this_filing.set_result(self.whole_filing_data)
            this_filing.set_keyerrors(self.filing_keyerr_data)
            return this_filing
        else:
            return this_filing
    """


    def run_sked(self, object_id, sked, verbose=False):
        """
        sked is the proper name of the schedule:
        IRS990, IRS990EZ, IRS990PF, IRS990ScheduleA, etc.
        """
        self.whole_filing_data = []
        self.filing_keyerr_data = []
        this_filing = Filing(object_id)
        this_filing.process(verbose=verbose)
        this_version = this_filing.get_version()
        if this_version in ALLOWED_VERSIONSTRINGS or ( self.csv_format and this_version in CSV_ALLOWED_VERSIONSTRINGS ):
            this_version = this_filing.get_version()
            ein = this_filing.get_ein()
            sked_dict = this_filing.get_schedule(sked)
            self._run_schedule(sked, object_id, sked_dict, ein)

            this_filing.set_result(self.whole_filing_data)
            this_filing.set_keyerrors(self.filing_keyerr_data)
            return this_filing
        else:
            print("Filing version %s isn't supported for this operation" % this_version )
            return this_filing



from distutils.core import setup
import os

NAME = 'irsx'
HUMAN_NAME = 'irsx'
HERE = os.path.abspath(os.path.dirname(__file__))
version_ns = {}
with open(os.path.join(HERE, 'irs_reader', '_version.py')) as f:
    exec(f.read(), {}, version_ns)

setup(name=HUMAN_NAME,
      description = "Turn the IRS' versioned XML 990's into python objects \
        with original line number and description.",
      version = version_ns['__version__'],
      author = 'Jacob Fenton',
      author_email = 'jsfenfen@gmail.com',
      url = 'https://github.com/jsfenfen/990-xml-reader',
      license = 'MIT',
      setup_requires = ["setuptools", ],
      install_requires = ['requests', 'xmltodict', 'unicodecsv'],
      tests_require = ['nose', 'requests', 'xmltodict', 'unicodecsv', 'tox', 'tox-pyenv',],
      packages = ['irsx'],
      package_dir = {'irsx': 'irs_reader'},
      package_data = {'irsx': ['metadata/*.csv']},
      keywords = ['990', 'nonprofit', 'tax'],
      entry_points = {
          "console_scripts": ["irsx=irsx.irsx_cli:main",
                              "irsx_index=irsx.irsx_index_cli:main",
                              "irsx_retrieve=irsx.irsx_retrieve_cli:main"]
      },
      classifiers=[
          # How mature is this project? Common values are
          #   3 - Alpha
          #   4 - Beta
          #   5 - Production/Stable
          'Development Status :: 3 - Alpha',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',

        ],
      )



from .type_utils import dictType, orderedDictType, listType, \
    unicodeType, noneType, strType
from .flatten_utils import flatten
from .keyerror_utils import ignorable_keyerror
from .settings import LOG_KEY


class SkedDictReader(object):
    """
    We get an ordered dict back from xmltodict, but we want to "flatten" it
    into xpath-ed variables and repeated structures.
    Will also work on reading xmltodict that was previously turned into json
    """
    def __init__(
        self,
        standardizer,
        groups,
        object_id,
        ein,
        documentId=None,
        documentation=False,
        csv_format=False
    ):

        self.standardizer = standardizer
        self.object_id = object_id
        self.ein = ein
        self.documentId = documentId
        self.schedule_parts = {}         # allows one entry per filing
        self.repeating_groups = {}       # multiple per filing
        self.csv_format = csv_format     # Do we need to generate ordered csv
        self.for_csv_list = []           # keep record of elements, line by line
        self.groups = groups
        self.documentation = documentation
        self.variable_keyerrors = []    # record any unexpected variables
        self.group_keyerrors = []       # or unexpected groups

        if self.documentation and not self.standardizer.get_documentation_status():
            # Todo: split out documenter entirely so we don't have to do this
            raise Exception(
                "Standardizer must be initialized with the \
                documentation flag to load documentation"
            )

    def _get_table_start(self):
        """ prefill the columns we need for all tables """
        if self.documentation:
            standardized_table_start = {
                'object_id': {
                    'value': self.object_id,
                    'ordering': -1,
                    'line_number': 'NA',
                    'description': 'IRS-assigned object id',
                    'db_type': 'String(18)'
                },
                'ein': {
                    'value': self.ein,
                    'ordering': -2,
                    'line_number': 'NA',
                    'description': 'IRS employer id number',
                    'db_type': 'String(9)'
                }
            }
            if self.documentId:
                standardized_table_start['documentId'] = {
                    'value': self.documentId,
                    'description': 'Document ID',
                    'ordering': 0
                }
        else:
            standardized_table_start = {
                'object_id': self.object_id,
                'ein': self.ein
            }
            if self.documentId:
                standardized_table_start['documentId'] = self.documentId

        return standardized_table_start

    def _process_group(self, json_node, path, this_group):

        for node_index, node in enumerate(json_node):
            #print("_process_group %s " % (this_group['db_name']))
            this_node_type = type(node)
            flattened_list_item = None
            if this_node_type == unicodeType:
                #print("_pg: unicodeType %s ")
                flattened_list_item = {path: node}
            else:
                #print("_pg: NOT unicodeType")
                flattened_list_item = flatten(node, parent_key=path, sep='/')
            table_name = None
            standardized_group_dict = self._get_table_start()

            for xpath in flattened_list_item.keys():
                if '@' in xpath:
                    continue
                else:
                    xpath = xpath.replace("/#text", "")
                    value = flattened_list_item[xpath]

                    if self.csv_format:
                        this_var = {
                            'xpath':xpath,
                            'value':value,
                            'in_group':True,
                            'group_name':this_group['db_name'],
                            'group_index':node_index
                        }
                        self.for_csv_list.append(this_var)

                    try:
                        this_var_data = self.standardizer.get_var(xpath)
                    except KeyError:
                        if not ignorable_keyerror(xpath):
                            self.variable_keyerrors.append(
                                {'element_path':xpath}
                            )
                        continue
                    this_var_value = flattened_list_item[xpath]
                    this_var_name = this_var_data['db_name']
                    table_name = this_var_data['db_table']
                    if self.documentation:
                        result = {
                            'value': this_var_value,
                            'ordering': this_var_data['ordering'],
                            'line_number': this_var_data['line_number'],
                            'description': this_var_data['description'],
                            'db_type': this_var_data['db_type']
                        }
                        standardized_group_dict[this_var_name] = result

                    else:
                        standardized_group_dict[this_var_name] = this_var_value
            try:
                self.repeating_groups[table_name].append(standardized_group_dict)
            except KeyError:
                self.repeating_groups[table_name] = [standardized_group_dict]

    def _parse_json(self, json_node, parent_path=""):
        this_node_type = type(json_node)
        element_path = parent_path

        if this_node_type == listType:
            #print("List type %s" % element_path)

            this_group = None
            try:
                this_group = self.groups[element_path]
            except KeyError:
                self.group_keyerrors.append(
                    {'element_path':element_path}
                )
            self._process_group(json_node, parent_path, this_group)

        elif this_node_type == unicodeType:
            # but ignore it if is an @.
            if '@' in element_path:
                pass
            else:
                element_path = element_path.replace("/#text", "")
                try:
                    # is it a group?
                    this_group = self.groups[element_path]
                    self._process_group(
                        [{parent_path: json_node}],
                        '',
                        this_group
                    )

                except KeyError:

                    # It's not a group so it should be a variable we know about
                    
                    if self.csv_format:
                        this_var = {
                            'xpath':element_path,
                            'value':json_node,
                            'in_group':False,
                            'group_name':None,
                            'group_index':None
                        }
                        self.for_csv_list.append(this_var)

                    # It's not a group so it should be a variable we know about
                    try:
                        var_data = self.standardizer.get_var(element_path)
                        var_found = True

                    except KeyError:
                        # pass through for some common key errors
                        # [ TODO: FIX THE KEYERRORS! ]
                        if not ignorable_keyerror(element_path):
                            self.variable_keyerrors.append(
                                {'element_path':element_path}
                            )
                        var_found = False

                    if var_found:

                        table_name = var_data['db_table']
                        var_name = var_data['db_name']

                        result = json_node
                        if self.documentation:
                            result = {
                                'value': json_node,
                                'ordering': var_data['ordering'],
                                'line_number': var_data['line_number'],
                                'description': var_data['description'],
                                'db_type': var_data['db_type']
                            }

                        try:
                            self.schedule_parts[table_name][var_name] = result
                        except KeyError:
                            self.schedule_parts[table_name] = self._get_table_start()
                            self.schedule_parts[table_name][var_name] = result


        elif this_node_type == orderedDictType or this_node_type == dictType:

            try:
                # is it a singleton group?
                this_group = self.groups[element_path]
                self._process_group([{parent_path: json_node}], '', this_group)

            except KeyError:
                keys = json_node.keys()
                for key in keys:
                    new_path = parent_path + "/" + key
                    self._parse_json(json_node[key], parent_path=new_path)

        elif this_node_type == noneType:
            pass

        elif this_node_type == strType:
            msg = "String '%s'" % json_node
            #self.logging.debug(msg)
        else:
            raise Exception("Unhandled type: %s" % (type(json_node)))

    def parse(self, raw_ordered_dict, parent_path=""):
        self._parse_json(raw_ordered_dict, parent_path=parent_path)
        return ({
            'schedule_parts': self.schedule_parts,
            'groups': self.repeating_groups,
            'csv_line_array':self.for_csv_list,    # This is empty if not csv
            'keyerrors':self.variable_keyerrors,
            'group_keyerrors':self.group_keyerrors
        })
